import os
import re

import yaml
from vcr import VCR
from vcr.request import Request
from vcr.serialize import CASSETTE_FORMAT_VERSION, _warn_about_old_cassette_format, _looks_like_an_old_cassette
from vcr.serializers import compat

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


class BinarySerializer(object):

    @classmethod
    def deserialize(cls, cassette_string):
        try:
            data = yaml.load(cassette_string, Loader=Loader)
        except (ImportError, yaml.constructor.ConstructorError):
            _warn_about_old_cassette_format()
        if _looks_like_an_old_cassette(data):
            _warn_about_old_cassette_format()

        requests = [Request._from_dict(r["request"]) for r in data["interactions"]]
        responses = [compat.convert_to_bytes(r["response"]) for r in data["interactions"] if r.get('response')]
        return requests, responses

    @classmethod
    def serialize(cls, cassette_dict):
        interactions = [
            {
                "id": str(i),
                "status": '',
                "request": compat.convert_to_unicode(r[0]._to_dict()),
                "response": compat.convert_to_unicode(r[1]),
            }
            for i, r in enumerate(zip(cassette_dict["requests"], cassette_dict["responses"]))
        ]
        data = {"version": CASSETTE_FORMAT_VERSION, "interactions": interactions}
        return yaml.dump(data, Dumper=Dumper)


class CustomPersister:
    @classmethod
    def load_cassette(cls, cassette_path, serializer=BinarySerializer):
        try:
            with open(cassette_path) as f:
                cassette_content = f.read()
        except OSError:
            raise ValueError("Cassette not found.")
        cassette = serializer.deserialize(cassette_content)
        return cassette

    @staticmethod
    def save_cassette(cassette_path, cassette_dict,
                      serializer=BinarySerializer):
        data = serializer.serialize(cassette_dict)
        dirname, filename = os.path.split(cassette_path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(cassette_path, "wb") as f:
            #f.write(data)
            f.write(data.encode('utf-8'))


def filter_cassette(
        interactions, id_=None, status=None, uri=None, method=None
):

    filters = []

    def id_filter(item):
        return item["id"] == id_

    def status_filter(item):
        return item["status"].upper() == status.upper()

    def uri_filter(item):
        return bool(re.search(uri, item["request"]["uri"]))

    def method_filter(item):
        return bool(re.search(method, item["request"]["method"]))

    if id_ is not None:
        filters.append(id_filter)

    if status is not None:
        filters.append(status_filter)

    if uri is not None:
        filters.append(uri_filter)

    if method is not None:
        filters.append(method_filter)

    def is_match(interaction):
        return all(filter_(interaction) for filter_ in filters)

    return filter(is_match, interactions)


d_vcr = VCR(record_mode='all')
d_vcr.register_serializer(name='binary', serializer=BinarySerializer())
d_vcr.serializer = 'binary'
d_vcr.register_persister(CustomPersister)



import json
from urllib.parse import urlparse

import requests
import attr
from requests.cookies import RequestsCookieJar
from requests.structures import CaseInsensitiveDict

from response_differ import cheks
from response_differ.cassettes import filter_cassette, d_vcr


@attr.s(slots=True)
class Replayed:
    interaction = attr.ib()
    response = attr.ib()
    responses = []
    config = attr.ib(default={})
    errors = []


def get_prepared_request(data):
    prepared = requests.PreparedRequest()
    prepared.method = data["method"]
    prepared.url = data["uri"]
    prepared._cookies = RequestsCookieJar()
    prepared.body = data["body"]
    prepared.headers = CaseInsensitiveDict([('Accept', '*/*')])
    if data.get('headers'):
        prepared.headers = CaseInsensitiveDict([(key, value[0]) for key, value in data["headers"].items()])
    #if data.get('headers'):
    #    prepared.headers = CaseInsensitiveDict(data["headers"])
    return prepared


def store_responses(replayed):
    if not replayed.interaction.get('response'):
        raise Exception("Not response. не передавайте флаг дифф")
    Replayed.responses.append(
        {
            'uri': urlparse(replayed.interaction['request']['uri']).path,
            'old': json.loads(replayed.interaction['response']['body']['string']),
            'new': replayed.response.json()
        })


def replay(cassette, cassette_path, status=None, uri=None, diff=None):
    #session = requests.Session()
    for interaction in filter_cassette(cassette["interactions"], status, uri):
        request = interaction["request"]
        if not request.get('headers'):
           request['headers'] = {}
        else:
            request['headers'] = CaseInsensitiveDict([(key, value[0]) for key, value in request["headers"].items()])
        #request = get_prepared_request(interaction["request"])
        with d_vcr.use_cassette(f'new_{cassette_path}'):
            response = requests.post(request['uri'], headers=request['headers'], data=request['body'])
            #response = session.send(request)
            replayed = Replayed(interaction, response)
            yield replayed
            cheks.check_status_code(replayed)
            diff and store_responses(replayed)



from hashlib import sha1


class CheckFailed(AssertionError):

    __module__ = "builtins"


CACHE = {}


def get_exception(name):
    if name in CACHE:
        exception_class = CACHE[name]
    else:
        exception_class = type(name, (CheckFailed,), {})
        exception_class.__qualname__ = CheckFailed.__name__
        exception_class.__name__ = CheckFailed.__name__
        CACHE[name] = exception_class
    return exception_class


def _get_hashed_exception(prefix, message):
    messages_digest = sha1(message.encode("utf-8")).hexdigest()
    name = f"{prefix}{messages_digest}"
    return get_exception(name)


def get_grouped_exception(*exceptions: AssertionError):
    messages = [exception.args[0] for exception in exceptions]
    message = "".join(messages)
    return _get_hashed_exception("GroupedException", f"{message}")


def get_status_code_error(status_code):
    name = f"StatusCodeError{status_code}"
    return get_exception(name)


def get_response_type_error(expected, received):
    name = f"SchemaValidationError{expected}_{received}"
    return get_exception(name)


def get_malformed_media_type_error(media_type):
    name = f"MalformedMediaType{media_type}"
    return get_exception(name)


def get_missing_content_type_error():
    return get_exception("MissingContentTypeError")


def get_response_parsing_error(exception):
    return _get_hashed_exception("ResponseParsingError", str(exception))


def get_headers_error(message):
    return _get_hashed_exception("MissingHeadersError", message)



import copy

from deepdiff import DeepDiff
from nested_lookup import nested_delete

#from response_differ import play, get_grouped_exception
from . import play
from .exceptions import get_grouped_exception


def check_status_code(replayed):
    if not replayed.response.ok:
        message = f"URL:{replayed.response.url}\n " \
                  f"STATUS CODE: {replayed.response.status_code}"
        play.Replayed.errors.append(AssertionError(message))


def check_diff_responses():
    for i in play.Replayed.responses:
        paths = play.Replayed.config['path']
        if i['uri'] in paths:
            old = i['old']
            new = i['new']
            if paths[i['uri']]:
                old = _filter(old, filters=paths[i['uri']])
                new = _filter(new, filters=paths[i['uri']])
            result = DeepDiff(old, new, ignore_order=True)
            print(result)
            if result.get('values_changed'):
                message = f"ERROR {result}"
                play.Replayed.errors.append(AssertionError(message))
                exception_cls = get_grouped_exception(*play.Replayed.errors)
                raise exception_cls(play.Replayed.errors)


def _filter(resp, filters):
    filtered_resp = copy.deepcopy(resp)
    for f in filters:
        filtered_resp = _filter(nested_delete(filtered_resp, f), filters=filters[1::])
    return filtered_resp



import json

import click
import yaml

from ..cheks import check_diff_responses
from ..exceptions import get_grouped_exception
from ..play import Replayed, replay


def bold(message):
    return click.style(message, bold=True)


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option()
def differ():
    """Command line tool for testing your web application"""


@differ.command()
@click.argument("cassette_path", type=click.Path(exists=True))
@click.option("--uri", help="A regexp that filters interactions by their request URI.", required=False, type=str)
@click.option("--diff", '-d',  help="Comparing new response with the old", required=False, type=bool)
@click.option("--config", '-c',  help="Comparing new response with the old", required=False, type=click.Path())
def run(cassette_path, status=None, uri=None, diff=False, config=None):
    click.secho(f"{bold('Replaying cassette')}: {cassette_path}")
    with open(cassette_path) as fd:
        cassette = yaml.load(fd, Loader=yaml.SafeLoader)
    click.secho(f"{bold('Total interactions')}: {len(cassette['interactions'])}\n")
    for replayed in replay(cassette, cassette_path, status=status, uri=uri, diff=diff):
        click.secho(f"  {bold('ID')}              | {replayed.interaction['id']}")
        click.secho(f"  {bold('URI')}             | {replayed.interaction['request']['uri']}")
        if replayed.interaction.get('response'):
            click.secho(f"  {bold('Old status code')} | {replayed.interaction['response']['status']['code']}")
        if hasattr(replayed, 'response'):
            click.secho(f"  {bold('New status code')} | {replayed.response.status_code}")
        click.secho(f"  {bold('Old request')}     | {replayed.interaction['request']}")
        if hasattr(replayed, 'response'):
            click.secho(f"  {bold('New request')}     | {replayed.response.request.body}\n")
    if Replayed.errors:
        exception_cls = get_grouped_exception(*Replayed.errors)
        raise exception_cls(Replayed.errors)
    if diff:
        with open(config) as fd2:
            Replayed.config =yaml.load(fd2, Loader=yaml.SafeLoader)
        check_diff_responses()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['response_differ', 'response_differ.response_diff']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'click>=7.1.2,<8.0.0',
 'deepdiff>=5.3.0,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'vcrpy>=4.1.1,<5.0.0']

entry_points = \
{'console_scripts': ['differ = response_differ.response_diff.cli:differ']}

setup_kwargs = {
    'name': 'response-differ',
    'version': '0.1.19',
    'description': '',
    'long_description': None,
    'author': 'vkytepov',
    'author_email': 'vkytepov@live.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
